//**************************************************************************
//**
//**	##   ##    ##    ##   ##   ####     ####   ###     ###
//**	##   ##  ##  ##  ##   ##  ##  ##   ##  ##  ####   ####
//**	 ## ##  ##    ##  ## ##  ##    ## ##    ## ## ## ## ##
//**	 ## ##  ########  ## ##  ##    ## ##    ## ##  ###  ##
//**	  ###   ##    ##   ###    ##  ##   ##  ##  ##       ##
//**	   #    ##    ##    #      ####     ####   ##       ##
//**
//**	Copyright (C) 1999-2000 Jnis Legzdi
//**
//**	This program is free software; you can redistribute it and/or
//**  modify it under the terms of the GNU General Public License
//**  as published by the Free Software Foundation; either version 2
//**  of the License, or (at your option) any later version.
//**
//**	This program is distributed in the hope that it will be useful,
//**  but WITHOUT ANY WARRANTY; without even the implied warranty of
//**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//**  GNU General Public License for more details.
//**	
//**************************************************************************

#define USE_GWND_AMMO_1	1
#define USE_GWND_AMMO_2 1
#define USE_CBOW_AMMO_1 1
#define USE_CBOW_AMMO_2 1
#define USE_BLSR_AMMO_1 1
#define USE_BLSR_AMMO_2 5
#define USE_SKRD_AMMO_1 1
#define USE_SKRD_AMMO_2 5
#define USE_PHRD_AMMO_1 1
#define USE_PHRD_AMMO_2 1
#define USE_MACE_AMMO_1 1
#define USE_MACE_AMMO_2 5

#define MAX_MACE_SPOTS	8

int WeaponAmmoUsePL1[NUMWEAPONS] =
{
	0,					// staff
	USE_GWND_AMMO_1,	// gold wand
	USE_CBOW_AMMO_1,	// crossbow
	USE_BLSR_AMMO_1,	// blaster
	USE_SKRD_AMMO_1,	// skull rod
	USE_PHRD_AMMO_1,	// phoenix rod
	USE_MACE_AMMO_1,	// mace
	0,					// gauntlets
	0					// beak
};

int WeaponAmmoUsePL2[NUMWEAPONS] =
{
	0,					// staff
	USE_GWND_AMMO_2,	// gold wand
	USE_CBOW_AMMO_2,	// crossbow
	USE_BLSR_AMMO_2,	// blaster
	USE_SKRD_AMMO_2,	// skull rod
	USE_PHRD_AMMO_2,	// phoenix rod
	USE_MACE_AMMO_2,	// mace
	0,					// gauntlets
	0					// beak
};

int 			MaceSpotCount;
float			MaceSpot_x[MAX_MACE_SPOTS];
float			MaceSpot_y[MAX_MACE_SPOTS];

//****************************************************************************
//
//	WEAPON UTILITES
//
//****************************************************************************

//===========================================================================
//
//	OpenWeapons
//
//	Called at level load before things are loaded.
//
//===========================================================================

void OpenWeapons(void)
{
	MaceSpotCount = 0;
}

//===========================================================================
//
//	AddMaceSpot
//
//===========================================================================

void AddMaceSpot(float x, float y)
{
	if (MaceSpotCount == MAX_MACE_SPOTS)
	{
		Error("Too many mace spots.");
	}
	MaceSpot_x[MaceSpotCount] = x;
	MaceSpot_y[MaceSpotCount] = y;
	MaceSpotCount++;
}

//===========================================================================
//
//  RepositionMace
//
//	Chooses the next spot to place the mace.
//
//===========================================================================

void RepositionMace(mobj_t *mo)
{
	int 		spot;

	spot = P_Random() % MaceSpotCount;
	P_UnsetThingPosition(mo);
	mo->origin.x = MaceSpot_x[spot];
	mo->origin.y = MaceSpot_y[spot];
	P_SetThingPosition(mo);
	mo->origin.z = mo->floorz;
}

//===========================================================================
//
//	CloseWeapons
//
//	Called at level load after things are loaded.
//
//===========================================================================

void CloseWeapons(void)
{
	int spot;

	if (!MaceSpotCount)
	{
		// No maces placed
		return;
	}
	if (!deathmatch && P_Random() < 64)
	{
		// Sometimes doesn't show up if not in deathmatch
		return;
	}
	spot = P_Random() % MaceSpotCount;
	P_SpawnMobj(MaceSpot_x[spot], MaceSpot_y[spot], ONFLOORZ, MT_WMACE);
}

//===========================================================================
//
//  BringUpWeapon
//
//	Starts bringing the pending weapon up from the bottom of the screen.
//
//===========================================================================

void BringUpWeapon(player_t *player)
{
	player->pspriteSY = PSpriteSY[player->readyweapon];
	weaponinfo[player->readyweapon](player);
	if (player->readyweapon == wp_gauntlets)
	{
		StartSoundName(player->mo, "GauntletActive");
	}
	player->pendingweapon = wp_nochange;
	player->psprites[ps_weapon].sy = WEAPONBOTTOM;
	if (player->powers[pw_weaponlevel2])
	{
		SetPsprite(player, ps_weapon, player->w2_upstate);
	}
	else
	{
		SetPsprite(player, ps_weapon, player->w_upstate);
	}
}

//===========================================================================
//
//	DropWeapon
//
//	The player died, so put the weapon away.
//
//===========================================================================

void DropWeapon(player_t *player)
{
	if (player->powers[pw_weaponlevel2])
	{
		SetPsprite(player, ps_weapon, player->w2_downstate);
	}
	else
	{
		SetPsprite(player, ps_weapon, player->w_downstate);
	}
}

//===========================================================================
//
//	SetupPsprites
//
//	Called at start of level for each player
//
//===========================================================================

void SetupPsprites(player_t *player)
{
	int 	i;

	// Remove all psprites
	for (i = 0; i < NUMPSPRITES; i++)
	{
		player->psprites[i].statenum = 0;
	}
	// Spawn the ready weapon
	BringUpWeapon(player);
}

//===========================================================================
//
//	MovePsprites
//
//	Called every tic by player thinking routine
//
//===========================================================================

void MovePsprites(player_t *player)
{
	int 		i;
	pspdef_t 	*psp;

	for(i = 0; i < NUMPSPRITES; i++)
	{
		psp = &player->psprites[i];
		if (psp->statenum) // a null state means not active
		{
			// drop tic count and possibly change state
			if (psp->time != -1.0)	// a -1 tic count never changes
			{
				psp->time -= frametime;
				if (psp->time <= 0.0)
				{
					psp->time = 0.0;
					SetPsprite(player, i, psp->nextstate);
				}
			}
		}
	}
	player->psprites[ps_flash].sx = player->psprites[ps_weapon].sx;
	player->psprites[ps_flash].sy = player->psprites[ps_weapon].sy;
}

//===========================================================================
//
//  CheckAmmo
//
//	Returns true if there is enough ammo to shoot.  If not, selects the
// next weapon to use.
//
//===========================================================================

boolean CheckAmmo(player_t *player)
{
	int ammo;
	int *ammoUse;
	int count;

	ammo = player->w_ammo;
	if(player->powers[pw_weaponlevel2] && !deathmatch)
	{
		ammoUse = &WeaponAmmoUsePL2[0];
	}
	else
	{
		ammoUse = &WeaponAmmoUsePL1[0];
	}
	count = ammoUse[player->readyweapon];
	if(ammo == am_noammo || player->ammo[ammo] >= count)
	{
		return true;
	}
	// out of ammo, pick a weapon to change to
	do
	{
		if(player->weaponowned[wp_skullrod]
			&& player->ammo[am_skullrod] > ammoUse[wp_skullrod])
		{
			player->pendingweapon = wp_skullrod;
		}
		else if(player->weaponowned[wp_blaster]
			&& player->ammo[am_blaster] > ammoUse[wp_blaster])
		{
			player->pendingweapon = wp_blaster;
		}
		else if(player->weaponowned[wp_crossbow]
			&& player->ammo[am_crossbow] > ammoUse[wp_crossbow])
		{
			player->pendingweapon = wp_crossbow;
		}
		else if(player->weaponowned[wp_mace]
			&& player->ammo[am_mace] > ammoUse[wp_mace])
		{
			player->pendingweapon = wp_mace;
		}
		else if(player->ammo[am_goldwand] > ammoUse[wp_goldwand])
		{
			player->pendingweapon = wp_goldwand;
		}
		else if(player->weaponowned[wp_gauntlets])
		{
			player->pendingweapon = wp_gauntlets;
		}
		else if(player->weaponowned[wp_phoenixrod]
			&& player->ammo[am_phoenixrod] > ammoUse[wp_phoenixrod])
		{
			player->pendingweapon = wp_phoenixrod;
		}
		else
		{
			player->pendingweapon = wp_staff;
		}
	} while(player->pendingweapon == wp_nochange);
	if(player->powers[pw_weaponlevel2])
	{
		SetPsprite(player, ps_weapon, player->w2_downstate);
	}
	else
	{
		SetPsprite(player, ps_weapon, player->w_downstate);
	}
	return false;
}

//===========================================================================
//
//  BulletSlope
//
//	Sets a slope so a near miss is at aproximately the height of the
// intended target
//
//===========================================================================

void BulletSlope(mobj_t *mo, TAVec *angles)
{
	//
	// see which target is to be aimed at
	//
	angles->pitch = mo->angles.pitch;
	angles->yaw = mo->angles.yaw;
	angles->roll = mo->angles.roll;

	AimLineAttack(mo, angles, 16.0 * 64.0);
	if (!linetarget)
	{
		angles->yaw += 1 << 26;
		AimLineAttack(mo, angles, 16.0 * 64.0);
		if (!linetarget)
		{
			angles->yaw -= 2 << 26;
			AimLineAttack(mo, angles, 16.0 * 64.0);
			if (!linetarget)
			{
				angles->yaw += 1 << 26;
			}
		}
	}
}

//===========================================================================
//
//  FireWeapon
//
//===========================================================================

void FireWeapon(player_t *player)
{
	int attackState;

	if (!CheckAmmo(player))
	{
		return;
	}
	SetMobjState(player->mo, S_PLAY_ATK2);
	if (player->powers[pw_weaponlevel2])
 	{
		attackState = player->refire ? player->w2_holdatkstate
			: player->w2_atkstate;
 	}
 	else
 	{
		attackState = player->refire ? player->w_holdatkstate
			: player->w_atkstate;
 	}
	SetPsprite(player, ps_weapon, attackState);
	P_NoiseAlert(player->mo, player->mo);
	if (player->readyweapon == wp_gauntlets && !player->refire)
	{
		// Play the sound for the initial gauntlet attack
		StartSoundName(player->mo, "GauntletUse");
	}
}

//===========================================================================
//
//  ActivateBeak
//
//===========================================================================

void ActivateBeak(player_t *player)
{
	player->pendingweapon = wp_nochange;
	player->readyweapon = wp_beak;
	weaponinfo[wp_beak](player);
	player->pspriteSY = PSpriteSY[player->readyweapon];
	player->psprites[ps_weapon].sy = WEAPONTOP;
	SetPsprite(player, ps_weapon, S_BEAKREADY);
}

//===========================================================================
//
//  PostChickenWeapon
//
//===========================================================================

void PostChickenWeapon(player_t *player, int weapon)
{
	if(weapon == wp_beak)
	{ // Should never happen
		weapon = wp_staff;
	}
	player->pendingweapon = wp_nochange;
	player->readyweapon = weapon;
	weaponinfo[weapon](player);
	player->pspriteSY = PSpriteSY[player->readyweapon];
	player->psprites[ps_weapon].sy = WEAPONBOTTOM;
	SetPsprite(player, ps_weapon, player->w_upstate);
}

//===========================================================================
//
//  UpdateBeak
//
//===========================================================================

void UpdateBeak(player_t *player, pspdef_t *psp)
{
	psp->sy = WEAPONTOP + itof(player->chickenPeck) / 2.0;
}

//===========================================================================
//
//	WeaponLevel2Fade
//
//===========================================================================

void WeaponLevel2Fade(player_t* player)
{
	if ((player->readyweapon == wp_phoenixrod) &&
		(player->psprites[ps_weapon].statenum != S_PHOENIXREADY) &&
		(player->psprites[ps_weapon].statenum != S_PHOENIXUP))
	{
		SetPsprite(player, ps_weapon, S_PHOENIXREADY);
		player->ammo[am_phoenixrod] -= USE_PHRD_AMMO_2;
		player->refire = 0;
	}
	else if((player->readyweapon == wp_gauntlets)
		|| (player->readyweapon == wp_staff))
	{
		player->pendingweapon = player->readyweapon;
	}
}

//****************************************************************************
//
//	WEAPON ACTIONS
//
//****************************************************************************

//===========================================================================
//
//	A_WeaponReady
//
//	The player can fire the weapon or change to another weapon at this time.
//
//===========================================================================

void A_WeaponReady(mobj_t *actor)
{
	player_t	*player = actor->player;
	pspdef_t	*psp = &player->psprites[ps_weapon];
	int 		angle;

	if (player->chickenTics)
	{
		// Change to the chicken beak
		ActivateBeak(player);
		return;
	}
	// Change player from attack state
	if (player->mo->statenum == S_PLAY_ATK1 ||
		player->mo->statenum == S_PLAY_ATK2)
	{
		SetMobjState(player->mo, S_PLAY);
	}
	// Check for staff PL2 active sound
	if ((player->readyweapon == wp_staff) &&
		(psp->statenum == S_STAFFREADY2_1) &&
		P_Random() < 128)
	{
		StartSoundName(player->mo, "StaffActive");
	}
	// Put the weapon away if the player has a pending weapon or has
	// died.
	if(player->pendingweapon != wp_nochange || !player->health)
	{
		if(player->powers[pw_weaponlevel2])
		{
			SetPsprite(player, ps_weapon, player->w2_downstate);
		}
		else
		{
			SetPsprite(player, ps_weapon, player->w_downstate);
		}
		return;
	}

	// Check for fire.  The phoenix rod does not auto fire.
	if (player->buttons & BT_ATTACK)
	{
		if (!player->attackdown || (player->readyweapon != wp_phoenixrod))
		{
			player->attackdown = true;
			FireWeapon(player);
			return;
		}
	}
	else
	{
		player->attackdown = false;
	}

	// Bob the weapon based on movement speed.
	angle = (ANG90 >> 4) * level->tictime;
	psp->sx = 1.0 + player->bob * cos(angle);
	angle &= 0x7fffffff;
	psp->sy = WEAPONTOP + player->bob * sin(angle);
}

//===========================================================================
//
//	A_BeakReady
//
//===========================================================================

void A_BeakReady(mobj_t *actor)
{
	player_t	*player = actor->player;
	if (player->buttons & BT_ATTACK)
	{
		// Chicken beak attack
		player->attackdown = true;
		SetMobjState(player->mo, S_CHICPLAY_ATK1);
		if (player->powers[pw_weaponlevel2])
		{
			SetPsprite(player, ps_weapon, S_BEAKATK2_1);
		}
		else
		{
			SetPsprite(player, ps_weapon, S_BEAKATK1_1);
		}
		P_NoiseAlert(player->mo, player->mo);
	}
	else
	{
		if (player->mo->statenum == S_CHICPLAY_ATK1)
		{ // Take out of attack state
			SetMobjState(player->mo, S_CHICPLAY);
		}
		player->attackdown = false;
	}
}

//===========================================================================
//
//	A_Lower
//
//===========================================================================

void A_Lower(mobj_t *actor)
{
	player_t	*player = actor->player;
	pspdef_t	*psp = &player->psprites[ps_weapon];
	if (player->chickenTics)
	{
		psp->sy = WEAPONBOTTOM;
	}
	else
	{
		psp->sy += LOWERSPEED;
	}
	if (psp->sy < WEAPONBOTTOM)
	{
		// Not lowered all the way yet
		return;
	}
	if (player->playerstate == PST_DEAD)
	{
		// Player is dead, so don't bring up a pending weapon
		psp->sy = WEAPONBOTTOM;
		return;
	}
	if (!player->health)
	{
		// Player is dead, so keep the weapon off screen
		SetPsprite(player,  ps_weapon, S_NULL);
		return;
	}
	player->readyweapon = player->pendingweapon;
	BringUpWeapon(player);
}

//===========================================================================
//
//	A_Raise
//
//===========================================================================

void A_Raise(mobj_t *actor)
{
	player_t	*player = actor->player;
	pspdef_t	*psp = &player->psprites[ps_weapon];
	psp->sy -= RAISESPEED;
	if (psp->sy > WEAPONTOP)
	{
		// Not raised all the way yet
		return;
	}
	psp->sy = WEAPONTOP;
	if (player->powers[pw_weaponlevel2])
	{
		SetPsprite(player, ps_weapon, player->w2_readystate);
	}
	else
	{
		SetPsprite(player, ps_weapon, player->w_readystate);
	}
}

//===========================================================================
//
//	A_BeakRaise
//
//===========================================================================

void A_BeakRaise(mobj_t *actor)
{
	player_t	*player = actor->player;
	pspdef_t	*psp = &player->psprites[ps_weapon];
	psp->sy = WEAPONTOP;
	SetPsprite(player, ps_weapon, player->w_readystate);
}

//===========================================================================
//
//	A_ReFire
//
// 	The player can re fire the weapon without lowering it entirely.
//
//===========================================================================

void A_ReFire(mobj_t *actor)
{
	player_t	*player = actor->player;
	if ((player->buttons & BT_ATTACK)
		&& player->pendingweapon == wp_nochange && player->health)
	{
		player->refire++;
		FireWeapon(player);
	}
	else
	{
		player->refire = 0;
		CheckAmmo(player);
	}
}

//===========================================================================
//
//	A_Light0
//
//===========================================================================

void A_Light0(mobj_t *actor)
{
	player_t	*player = actor->player;
	player->extralight = 0;
}

//****************************************************************************
//
//	WEAPON ATTACKS
//
//****************************************************************************

//===========================================================================
//
//	A_StaffAttackPL1
//
//===========================================================================

void A_StaffAttackPL1(mobj_t *actor)
{
	player_t	*player = actor->player;
	TAVec	angles;
	int 	damage;
	TVec	vforward;

	damage = 5 + (P_Random() & 15);
	avcpy(player->mo->angles, angles);
	angles.yaw += (P_Random() - P_Random()) << 18;
	AimLineAttack(player->mo, &angles, MELEERANGE);
	AngleVector(&angles, &vforward);
	PuffType = MT_STAFFPUFF;
	LineAttack(player->mo, &vforward, MELEERANGE, damage);
	if (linetarget)
	{
		//StartSoundName(player->mo, sfx_stfhit);
		// turn to face target
		player->mo->angles.yaw = atan2(linetarget->origin.y - player->mo->origin.y,
			linetarget->origin.x - player->mo->origin.x);
	}
}

//===========================================================================
//
//	A_StaffAttackPL2
//
//===========================================================================

void A_StaffAttackPL2(mobj_t *actor)
{
	player_t	*player = actor->player;
	TAVec	angles;
	int 	damage;
	TVec	vforward;

	// P_inter.c:DamageMobj() handles target momentums
	damage = 18 + (P_Random() & 63);

	avcpy(player->mo->angles, angles);
	angles.yaw += (P_Random() - P_Random()) << 18;
	AimLineAttack(player->mo, &angles, MELEERANGE);
	AngleVector(&angles, &vforward);
	PuffType = MT_STAFFPUFF2;
	LineAttack(player->mo, &vforward, MELEERANGE, damage);
	if (linetarget)
	{
		//StartSoundName(player->mo, sfx_stfpow);
		// turn to face target
		player->mo->angles.yaw = atan2(linetarget->origin.y - player->mo->origin.y,
			linetarget->origin.x - player->mo->origin.x);
		player->fixangle = 1;
	}
}

//===========================================================================
//
//	A_GauntletAttack
//
//===========================================================================

void A_GauntletAttack(mobj_t *actor)
{
	player_t	*player = actor->player;
	pspdef_t	*psp = &player->psprites[ps_weapon];
	angle_t		angle;
	int 		damage;
	float		slope;
	int 		randVal;
	float		dist;
	TAVec		angles;
	TVec		vforward;

	psp->sx = Random() * 4.0 - 2.0;
	psp->sy = WEAPONTOP + Random() * 4.0;
	avcpy(player->mo->angles, angles);
	if (player->powers[pw_weaponlevel2])
	{
		damage = HITDICE(2);
		dist = 4.0 * MELEERANGE;
		angles.yaw += (P_Random()-P_Random())<<17;
		PuffType = MT_GAUNTLETPUFF2;
	}
	else
	{
		damage = HITDICE(2);
		dist = MELEERANGE + 0.00001;
		angles.yaw += (P_Random()-P_Random())<<18;
		PuffType = MT_GAUNTLETPUFF1;
	}
	AimLineAttack(player->mo, &angles, dist);
	AngleVector(&angles, &vforward);
	LineAttack(player->mo, &vforward, dist, damage);
	if (!linetarget)
	{
		if (P_Random() > 64)
		{
			player->extralight = !player->extralight;
		}
		StartSoundName(player->mo, "GauntletFull");
		return;
	}
	randVal = P_Random();
	if (randVal < 64)
	{
		player->extralight = 0;
	}
	else if (randVal < 160)
	{
		player->extralight = 1;
	}
	else
	{
		player->extralight = 2;
	}
	if (player->powers[pw_weaponlevel2])
	{
		P_GiveBody(player, damage >> 1);
		StartSoundName(player->mo, "GauntletHit2");
	}
	else
	{
		StartSoundName(player->mo, "GauntletHit");
	}
	// turn to face target
	angle = atan2(linetarget->origin.y - player->mo->origin.y,
		linetarget->origin.x - player->mo->origin.x);
	if (angle-player->mo->angles.yaw > ANG180)
	{
		if (angle - player->mo->angles.yaw < -ANG90 / 20)
			player->mo->angles.yaw = angle + ANG90 / 21;
		else
			player->mo->angles.yaw -= ANG90 / 20;
	}
	else
	{
		if(angle-player->mo->angles.yaw > ANG90 / 20)
			player->mo->angles.yaw = angle - ANG90 / 21;
		else
			player->mo->angles.yaw += ANG90 / 20;
	}
	player->mo->flags |= MF_JUSTATTACKED;
}

//===========================================================================
//
//	A_FireGoldWandPL1
//
//===========================================================================

void A_FireGoldWandPL1(mobj_t *actor)
{
	player_t	*player = actor->player;
	mobj_t		*mo;
	int			damage;
	TAVec		angles;
	TVec		vforward;

	mo = player->mo;
	player->ammo[am_goldwand] -= USE_GWND_AMMO_1;
	damage = 7 + (P_Random() & 7);
	BulletSlope(mo, &angles);
	if (player->refire)
	{
		angles.yaw += (P_Random() - P_Random()) << 18;
	}
	AngleVector(&angles, &vforward);
	PuffType = MT_GOLDWANDPUFF1;
	LineAttack(mo, &vforward, MISSILERANGE, damage);
	StartSoundName(player->mo, "GoldWandFire");
}

//===========================================================================
//
//	A_FireGoldWandPL2
//
//===========================================================================

void A_FireGoldWandPL2(mobj_t *actor)
{
	player_t	*player = actor->player;
	int 		i;
	mobj_t*		mo;
	int 		damage;
	float		momz;
	TAVec		angles;
	TVec		vforward;

	mo = player->mo;
	player->ammo[am_goldwand] -=
		deathmatch ? USE_GWND_AMMO_1 : USE_GWND_AMMO_2;
	PuffType = MT_GOLDWANDPUFF2;
	BulletSlope(mo, &angles);
	SpawnMissileAngles(mo, MT_GOLDWANDFX2, angles.yaw - (ANG45 / 8), angles.pitch);
	SpawnMissileAngles(mo, MT_GOLDWANDFX2, angles.yaw + (ANG45 / 8), angles.pitch);
	angles.yaw -= (ANG45 / 8);
	for (i = 0; i < 5; i++)
	{
		damage = 1 + (P_Random() & 7);
		AngleVector(&angles, &vforward);
		LineAttack(mo, &vforward, MISSILERANGE, damage);
		angles.yaw += ((ANG45 / 8) * 2) / 4;
	}
	StartSoundName(player->mo, "GoldWandFire");
}

//===========================================================================
//
//	A_FireCrossbowPL1
//
//===========================================================================

void A_FireCrossbowPL1(mobj_t *actor)
{
	player_t	*player = actor->player;
	mobj_t*	pmo;

	pmo = player->mo;
	player->ammo[am_crossbow] -= USE_CBOW_AMMO_1;
	SpawnPlayerMissile(pmo, MT_CRBOWFX1);
	SpawnPlayerMissileAngle(pmo, MT_CRBOWFX3, pmo->angles.yaw-(ANG45/10));
	SpawnPlayerMissileAngle(pmo, MT_CRBOWFX3, pmo->angles.yaw+(ANG45/10));
}

//===========================================================================
//
//	A_FireCrossbowPL2
//
//===========================================================================

void A_FireCrossbowPL2(mobj_t *actor)
{
	player_t	*player = actor->player;
	mobj_t*	pmo;

	pmo = player->mo;
	player->ammo[am_crossbow] -=
		deathmatch ? USE_CBOW_AMMO_1 : USE_CBOW_AMMO_2;
	SpawnPlayerMissile(pmo, MT_CRBOWFX2);
	SpawnPlayerMissileAngle(pmo, MT_CRBOWFX2, pmo->angles.yaw-(ANG45/10));
	SpawnPlayerMissileAngle(pmo, MT_CRBOWFX2, pmo->angles.yaw+(ANG45/10));
	SpawnPlayerMissileAngle(pmo, MT_CRBOWFX3, pmo->angles.yaw-(ANG45/5));
	SpawnPlayerMissileAngle(pmo, MT_CRBOWFX3, pmo->angles.yaw+(ANG45/5));
}

//===========================================================================
//
//	A_FireBlasterPL1
//
//===========================================================================

void A_FireBlasterPL1(mobj_t *actor)
{
	player_t	*player = actor->player;
	mobj_t*		mo;
	int 		damage;
    float		bulletslope;
	TAVec		angles;
	TVec		vforward;

	mo = player->mo;
	StartSoundName(mo, "GoldWandFire");
	player->ammo[am_blaster] -= USE_BLSR_AMMO_1;
	BulletSlope(mo, &angles);
	damage = HITDICE(4);
	if (player->refire)
	{
		angles.yaw += (P_Random() - P_Random()) << 18;
	}
	PuffType = MT_BLASTERPUFF1;
	AngleVector(&angles, &vforward);
	LineAttack(mo, &vforward, MISSILERANGE, damage);
	StartSoundName(player->mo, "BlasterFire");
}

//===========================================================================
//
//	A_FireBlasterPL2
//
//===========================================================================

void A_FireBlasterPL2(mobj_t *actor)
{
	player_t	*player = actor->player;
	mobj_t*	mo;

	player->ammo[am_blaster] -=
		deathmatch ? USE_BLSR_AMMO_1 : USE_BLSR_AMMO_2;
	mo = SpawnPlayerMissile(player->mo, MT_BLASTERFX1);
	if (mo)
	{
    	mo->blaster = true;
	}
	StartSoundName(player->mo, "BlasterFire");
}

//===========================================================================
//
//	A_FireSkullRodPL1
//
//===========================================================================

void A_FireSkullRodPL1(mobj_t *actor)
{
	player_t	*player = actor->player;
	mobj_t*	mo;

	if (player->ammo[am_skullrod] < USE_SKRD_AMMO_1)
	{
		return;
	}
	player->ammo[am_skullrod] -= USE_SKRD_AMMO_1;
	mo = SpawnPlayerMissile(player->mo, MT_HORNRODFX1);
	// Randomize the first frame
	if (mo && P_Random() > 128)
	{
		SetMobjState(mo, S_HRODFX1_2);
	}
}

//===========================================================================
//
// 	A_FireSkullRodPL2
//
// 	The special2 field holds the player number that shot the rain missile.
// The specialmobj1 field is used for the seeking routines, special1 as a
// counter for the sound looping.
//
//===========================================================================

void A_FireSkullRodPL2(mobj_t *actor)
{
	player_t	*player = actor->player;
	player->ammo[am_skullrod] -=
		deathmatch ? USE_SKRD_AMMO_1 : USE_SKRD_AMMO_2;
	SpawnPlayerMissile(player->mo, MT_HORNRODFX2);
	// Use MissileMobj instead of the return value from
	// SpawnPlayerMissile because we need to give info to the mobj
	// even if it exploded immediately.
	if (netgame)
	{
		// Multi-player game
		MissileMobj->special2 = P_GetPlayerNum(player);
	}
	else
	{
		// Always use red missiles in single player games
		MissileMobj->special2 = 2;
	}
	if (linetarget)
	{
		MissileMobj->specialmobj1 = linetarget;
	}
	StartSoundName(MissileMobj, "SkullRodFire2");
}

//===========================================================================
//
//	A_FirePhoenixPL1
//
//===========================================================================

void A_FirePhoenixPL1(mobj_t *actor)
{
	player_t	*player = actor->player;
	TVec		vforward;

	player->ammo[am_phoenixrod] -= USE_PHRD_AMMO_1;
	SpawnPlayerMissile(player->mo, MT_PHOENIXFX1);
	AngleVector(&player->mo->angles, &vforward);
	vma(player->mo->velocity, vforward, -140.0, player->mo->velocity);
}

//===========================================================================
//
//	A_InitPhoenixPL2
//
//===========================================================================

void A_InitPhoenixPL2(mobj_t *actor)
{
	player_t	*player = actor->player;
	player->flamecount = ftoi(FLAME_THROWER_TIME * 35.0);
}

//===========================================================================
//
//	A_FirePhoenixPL2
//
// 	Flame thrower effect.
//
//===========================================================================

void A_FirePhoenixPL2(mobj_t *actor)
{
	player_t	*player = actor->player;
	mobj_t*		mo;
	mobj_t*		pmo;
	TVec		dir;
	float		x, y, z;

	if (--player->flamecount == 0)
	{
		// Out of flame
		SetPsprite(player, ps_weapon, S_PHOENIXATK2_4);
		player->refire = 0;
		return;
	}
	pmo = player->mo;
	x = pmo->origin.x + (Random() - Random()) * 2.0;
	y = pmo->origin.y + (Random() - Random()) * 2.0;
	z = pmo->origin.z + 26.0 - sin(player->mo->angles.pitch);
	z -= pmo->floorclip;
	mo = P_SpawnMobj(x, y, z, MT_PHOENIXFX2);
	mo->target = pmo;
	mo->angles.yaw = pmo->angles.yaw;
	mo->angles.pitch = pmo->angles.pitch;
	AngleVector(&mo->angles, &dir);
	mo->velocity = pmo->velocity + mo->speed * dir;
	if (!player->refire || !(level->tictime % 38))
	{
		StartSoundName(player->mo, "PhoenixHit2");
	}	
	CheckMissileSpawn(mo);
}

//===========================================================================
//
//	A_ShutdownPhoenixPL2
//
//===========================================================================

void A_ShutdownPhoenixPL2(mobj_t *actor)
{
	player_t	*player = actor->player;
	player->ammo[am_phoenixrod] -= USE_PHRD_AMMO_2;
}

//===========================================================================
//
// PROC A_FireMacePL1B
//
//===========================================================================

void A_FireMacePL1B(mobj_t *actor)
{
	player_t	*player = actor->player;
	mobj_t*	pmo;
	mobj_t*	ball;
	angle_t	angle;

	if (player->ammo[am_mace] < USE_MACE_AMMO_1)
	{
		return;
	}
	player->ammo[am_mace] -= USE_MACE_AMMO_1;
	pmo = player->mo;
	ball = P_SpawnMobj(pmo->origin.x, pmo->origin.y, pmo->origin.z + 28.0
		- pmo->floorclip, MT_MACEFX2);
	ball->velocity.z = (2.0 + 173.0 * -tan(player->mo->angles.pitch) / 32.0) * 35.0;
	angle = pmo->angles.yaw;
	ball->target = pmo;
	ball->angles.yaw = angle;
	ball->origin.z += 173.0 * -tan(player->mo->angles.pitch) / 16.0;
	ball->velocity.x = pmo->velocity.x / 2.0 + ball->speed * cos(angle);
	ball->velocity.y = pmo->velocity.y / 2.0 + ball->speed * sin(angle);
	StartSoundName(ball, "MaceFire");
	CheckMissileSpawn(ball);
}

//===========================================================================
//
//	A_FireMacePL1
//
//===========================================================================

void A_FireMacePL1(mobj_t *actor)
{
	player_t	*player = actor->player;
	pspdef_t	*psp = &player->psprites[ps_weapon];
	mobj_t*	ball;

	if (P_Random() < 28)
	{
		A_FireMacePL1B(actor);
		return;
	}
	if (player->ammo[am_mace] < USE_MACE_AMMO_1)
	{
		return;
	}
	player->ammo[am_mace] -= USE_MACE_AMMO_1;
	psp->sx = Random() * 4.0 - 2.0;
	psp->sy = WEAPONTOP + Random() * 4.0;
	ball = SpawnPlayerMissileAngle(player->mo, MT_MACEFX1, player->mo->angles.yaw
		+(((P_Random()&7)-4)<<24));
	if (ball)
	{
		ball->special1 = 16; // tics till dropoff
	}
}

//===========================================================================
//
//	A_FireMacePL2
//
//===========================================================================

void A_FireMacePL2(mobj_t *actor)
{
	player_t	*player = actor->player;
	mobj_t*	mo;

	player->ammo[am_mace] -=
		deathmatch ? USE_MACE_AMMO_1 : USE_MACE_AMMO_2;
	mo = SpawnPlayerMissile(player->mo, MT_MACEFX4);
	if (mo)
	{
		mo->velocity.x += player->mo->velocity.x;
		mo->velocity.y += player->mo->velocity.y;
		mo->velocity.z = (2.0 + 173.0 * -tan(player->mo->angles.pitch) / 32.0) * 35.0;
		if (linetarget)
		{
			mo->specialmobj1 = linetarget;
		}
	}
	StartSoundName(player->mo, "MaceFire");
}

//===========================================================================
//
//	A_BeakAttackPL1
//
//===========================================================================

void A_BeakAttackPL1(mobj_t *actor)
{
	player_t	*player = actor->player;
	pspdef_t	*psp = &player->psprites[ps_weapon];
	int 		damage;
	TAVec		angles;
	TVec		vforward;

	damage = 1+(P_Random()&3);
	avcpy(player->mo->angles, angles);
	AimLineAttack(player->mo, &angles, MELEERANGE);
	PuffType = MT_BEAKPUFF;
	AngleVector(&angles, &vforward);
	LineAttack(player->mo, &vforward, MELEERANGE, damage);
	if (linetarget)
	{
		player->mo->angles.yaw = atan2(linetarget->origin.y - player->mo->origin.y,
			linetarget->origin.x - player->mo->origin.x);
		player->fixangle = 1;
	}
	StartSound(player->mo, GetSoundID("ChickenPeak1") + (P_Random() % 3));
	player->chickenPeck = 12;
	psp->time -= Random() * 0.25;
}

//===========================================================================
//
//	A_BeakAttackPL2
//
//===========================================================================

void A_BeakAttackPL2(mobj_t *actor)
{
	player_t	*player = actor->player;
	pspdef_t	*psp = &player->psprites[ps_weapon];
	int 		damage;
	TAVec		angles;
	TVec		vforward;

	damage = HITDICE(4);
	avcpy(player->mo->angles, angles);
	AimLineAttack(player->mo, &angles, MELEERANGE);
	PuffType = MT_BEAKPUFF;
	AngleVector(&angles, &vforward);
	LineAttack(player->mo, &vforward, MELEERANGE, damage);
	if (linetarget)
	{
		player->mo->angles.yaw = atan2(linetarget->origin.y - player->mo->origin.y,
			linetarget->origin.x - player->mo->origin.x);
		player->fixangle = 1;
	}
	StartSound(player->mo, GetSoundID("ChickenPeak1") + (P_Random() % 3));
	player->chickenPeck = 12;
	psp->time -= Random() * 0.125;
}

//****************************************************************************
//
//  WEAPON PROJECTILE ACTIONS
//
//****************************************************************************

//===========================================================================
//
//	A_BoltSpark
//
//===========================================================================

void A_BoltSpark(mobj_t* bolt)
{
	mobj_t*	spark;

	if (P_Random() > 50)
	{
		spark = P_SpawnMobj(bolt->origin.x, bolt->origin.y, bolt->origin.z, MT_CRBOWFX4);
		spark->origin.x += (Random() - Random()) * 4.0;
		spark->origin.y += (Random() - Random()) * 4.0;
	}
}

//===========================================================================
//
//	A_SpawnRippers
//
//===========================================================================

void A_SpawnRippers(mobj_t* actor)
{
	int 	i;
	angle_t	angle;
	mobj_t*	ripper;

	for (i = 0; i < 8; i++)
	{
		ripper = P_SpawnMobj(actor->origin.x, actor->origin.y, actor->origin.z, MT_RIPPER);
		angle = i * ANG45;
		ripper->target = actor->target;
		ripper->angles.yaw = angle;
		ripper->velocity.x = ripper->speed * cos(angle);
		ripper->velocity.y = ripper->speed * sin(angle);
		CheckMissileSpawn(ripper);
	}
}

//===========================================================================
//
//	A_SkullRodPL2Seek
//
//===========================================================================

void A_SkullRodPL2Seek(mobj_t* actor)
{
	P_SeekerMissile(actor, ANGLE_1 * 10, ANGLE_1 * 30);
}

//===========================================================================
//
//	A_AddPlayerRain
//
//===========================================================================

void A_AddPlayerRain(mobj_t* actor)
{
	int 		playerNum;
	player_t*	player;

	playerNum = netgame ? actor->special2 : 0;
	player = &players[playerNum];
	if (!player->active)
	{
		// Player left the game
		return;
	}
	if (player->health <= 0)
	{
		// Player is dead
		return;
	}
	if (player->rain1 && player->rain2)
	{
		// Terminate an active rain
		if (player->rain1->health < player->rain2->health)
		{
			if (player->rain1->health > 16)
			{
				player->rain1->health = 16;
			}
			player->rain1 = NULL;
		}
		else
		{
			if (player->rain2->health > 16)
			{
				player->rain2->health = 16;
			}
			player->rain2 = NULL;
		}
	}
	// Add rain mobj to list
	if (player->rain1)
	{
		player->rain2 = actor;
	}
	else
	{
		player->rain1 = actor;
	}
}

//===========================================================================
//
//	A_SkullRodStorm
//
//===========================================================================

void A_SkullRodStorm(mobj_t* actor)
{
	float		x;
	float		y;
	mobj_t*		mo;
	int 		playerNum;
	player_t*	player;

	if (actor->health-- == 0)
	{
		SetMobjState(actor, S_NULL);
		playerNum = netgame ? actor->special2 : 0;
		player = &players[playerNum];
		if (!player->active)
		{
			// Player left the game
			return;
		}
		if (player->health <= 0)
		{
			// Player is dead
			return;
		}
		if (player->rain1 == actor)
		{
			player->rain1 = NULL;
		}
		else if (player->rain2 == actor)
		{
			player->rain2 = NULL;
		}
		return;
	}
	if (P_Random() < 25)
	{
		// Fudge rain frequency
		return;
	}
	x = actor->origin.x + Random() * 128.0 - 64.0;
	y = actor->origin.y + Random() * 128.0 - 64.0;
	mo = P_SpawnMobj(x, y, ONCEILINGZ, MT_RAINPLR1 + actor->special2);
	mo->target = actor->target;
	mo->velocity.x = 0.00001; // Force collision detection
	mo->velocity.z = -mo->speed;
	mo->special2 = actor->special2; // Transfer player number
	CheckMissileSpawn(mo);
	if (!(actor->special1 & 31))
	{
		StartSoundName(actor, "SkullRodRain");
	}
	actor->special1++;
}

//===========================================================================
//
//	A_RainImpact
//
//===========================================================================

void A_RainImpact(mobj_t* actor)
{
	if (actor->origin.z > actor->floorz)
	{
		SetMobjState(actor, S_RAINAIRXPLR1_1 + actor->special2);
	}
	else if (P_Random() < 40)
	{
		P_HitFloor(actor);
	}
}

//===========================================================================
//
//	A_HideInCeiling
//
//===========================================================================

void A_HideInCeiling(mobj_t* actor)
{
	actor->origin.z = actor->ceilingz + 4.0;
}

//===========================================================================
//
//	A_PhoenixPuff
//
//===========================================================================

void A_PhoenixPuff(mobj_t* actor)
{
	mobj_t*	puff;
	angle_t	angle;

	P_SeekerMissile(actor, ANGLE_1 * 5, ANGLE_1 * 10);
	puff = P_SpawnMobj(actor->origin.x, actor->origin.y, actor->origin.z, MT_PHOENIXPUFF);
	angle = actor->angles.yaw + ANG90;
	puff->velocity.x = 1.3 * cos(angle) * 35.0;
	puff->velocity.y = 1.3 * sin(angle) * 35.0;
	puff->velocity.z = 0.0;
	puff = P_SpawnMobj(actor->origin.x, actor->origin.y, actor->origin.z, MT_PHOENIXPUFF);
	angle = actor->angles.yaw - ANG90;
	puff->velocity.x = 1.3 * cos(angle) * 35.0;
	puff->velocity.y = 1.3 * sin(angle) * 35.0;
	puff->velocity.z = 0.0;
}

//===========================================================================
//
//	A_FlameEnd
//
//===========================================================================

void A_FlameEnd(mobj_t* actor)
{
	actor->velocity.z += 1.5 * 35.0;
}

//===========================================================================
//
// 	A_FloatPuff
//
//===========================================================================

void A_FloatPuff(mobj_t* puff)
{
	puff->velocity.z += 1.8 * 35.0;
}

//===========================================================================
//
//	A_MacePL1Check
//
//===========================================================================

void A_MacePL1Check(mobj_t* ball)
{
	int		angle;

	if (ball->special1 == 0)
	{
		return;
	}
	ball->special1 -= 4;
	if (ball->special1 > 0)
	{
		return;
	}
	ball->special1 = 0;
	ball->flags2 |= MF2_LOGRAV;
	angle = ball->angles.yaw;
	ball->velocity.x = 7.0 * cos(angle) * 35.0;
	ball->velocity.y = 7.0 * sin(angle) * 35.0;
	ball->velocity.z -= ball->velocity.z / 2.0;
}

//===========================================================================
//
//	A_MaceBallImpact
//
//===========================================================================

void A_MaceBallImpact(mobj_t* ball)
{
	if ((ball->origin.z <= ball->floorz) && (P_HitFloor(ball) != FLOOR_SOLID))
	{
		// Landed in some sort of liquid
		P_RemoveMobj(ball);
		return;
	}
	if ((ball->health != MAGIC_JUNK) && (ball->origin.z <= ball->floorz) && ball->velocity.z)
	{
		// Bounce
		ball->health = MAGIC_JUNK;
		ball->velocity.z = ball->velocity.z * 192.0 / 256.0;
		ball->flags2 &= ~MF2_FLOORBOUNCE;
		SetMobjState(ball, ball->spawnstate);
		StartSoundName(ball, "Bounce");
	}
	else
	{
		// Explode
		ball->flags |= MF_NOGRAVITY;
		ball->flags2 &= ~MF2_LOGRAV;
		StartSoundName(ball, "MaceHit");
	}
}

//===========================================================================
//
//	A_MaceBallImpact2
//
//===========================================================================

void A_MaceBallImpact2(mobj_t* ball)
{
	mobj_t*	tiny;
	angle_t	angle;

	if ((ball->origin.z <= ball->floorz) && (P_HitFloor(ball) != FLOOR_SOLID))
	{
		// Landed in some sort of liquid
		P_RemoveMobj(ball);
		return;
	}
	if ((ball->origin.z != ball->floorz) || (ball->velocity.z < 2.0 * 35.0))
	{
		// Explode
		ball->velocity.x = ball->velocity.y = ball->velocity.z = 0.0;
		ball->flags |= MF_NOGRAVITY;
		ball->flags2 &= ~(MF2_LOGRAV|MF2_FLOORBOUNCE);
	}
	else
	{
		// Bounce
		ball->velocity.z = ball->velocity.z * 192.0 / 256.0;
		SetMobjState(ball, ball->spawnstate);

		tiny = P_SpawnMobj(ball->origin.x, ball->origin.y, ball->origin.z, MT_MACEFX3);
		angle = ball->angles.yaw+ANG90;
		tiny->target = ball->target;
		tiny->angles.yaw = angle;
		tiny->velocity.x = ball->velocity.x / 2.0 + (ball->velocity.z - 35.0) * cos(angle);
		tiny->velocity.y = ball->velocity.y / 2.0 + (ball->velocity.z - 35.0) * sin(angle);
		tiny->velocity.z = ball->velocity.z;
		CheckMissileSpawn(tiny);

		tiny = P_SpawnMobj(ball->origin.x, ball->origin.y, ball->origin.z, MT_MACEFX3);
		angle = ball->angles.yaw - ANG90;
		tiny->target = ball->target;
		tiny->angles.yaw = angle;
		tiny->velocity.x = ball->velocity.x / 2.0 + (ball->velocity.z - 35.0) * cos(angle);
		tiny->velocity.y = ball->velocity.y / 2.0 + (ball->velocity.z - 35.0) * sin(angle);
		tiny->velocity.z = ball->velocity.z;
		CheckMissileSpawn(tiny);
	}
}

//===========================================================================
//
//	A_DeathBallImpact
//
//===========================================================================

void A_DeathBallImpact(mobj_t* ball)
{
	int 	i;
	mobj_t*	target;
	angle_t	angle;
	int		newAngle;
	TAVec	angles;

	if ((ball->origin.z <= ball->floorz) && (P_HitFloor(ball) != FLOOR_SOLID))
	{
		// Landed in some sort of liquid
		P_RemoveMobj(ball);
		return;
	}
	if ((ball->origin.z <= ball->floorz) && ball->velocity.z)
	{
		// Bounce
		newAngle = false;
		target = ball->specialmobj1;
		if (target)
		{
			if (!(target->flags&MF_SHOOTABLE))
			{
				// Target died
				ball->specialmobj1 = NULL;
			}
			else
			{
				// Seek
				angle = atan2(target->origin.y - ball->origin.y,
					target->origin.x - ball->origin.x);
				newAngle = true;
			}
		}
		else
		{
			// Find new target
			avset(angles, 0, 0, 0);
			for (i = 0; i < 16; i++)
			{
				AimLineAttack(ball, &angles, 10.0 * 64.0);
				if (linetarget && ball->target != linetarget)
				{
					ball->specialmobj1 = linetarget;
					angle = atan2(linetarget->origin.y - ball->origin.y,
						linetarget->origin.x - ball->origin.x);
					newAngle = true;
					break;
				}
				angle += ANG45 / 2;
			}
		}
		if (newAngle)
		{
			ball->angles.yaw = angle;
			ball->velocity.x = ball->speed * cos(angle);
			ball->velocity.y = ball->speed * sin(angle);
		}
		SetMobjState(ball, ball->spawnstate);
		StartSoundName(ball, "PlatformStop");
	}
	else
	{ // Explode
		ball->flags |= MF_NOGRAVITY;
		ball->flags2 &= ~MF2_LOGRAV;
		StartSoundName(ball, "PhoenixHit");
	}
}


