//**************************************************************************
//**
//**	##   ##    ##    ##   ##   ####     ####   ###     ###
//**	##   ##  ##  ##  ##   ##  ##  ##   ##  ##  ####   ####
//**	 ## ##  ##    ##  ## ##  ##    ## ##    ## ## ## ## ##
//**	 ## ##  ########  ## ##  ##    ## ##    ## ##  ###  ##
//**	  ###   ##    ##   ###    ##  ##   ##  ##  ##       ##
//**	   #    ##    ##    #      ####     ####   ##       ##
//**
//**	Copyright (C) 1999-2000 Jnis Legzdi
//**
//**	This program is free software; you can redistribute it and/or
//**  modify it under the terms of the GNU General Public License
//**  as published by the Free Software Foundation; either version 2
//**  of the License, or (at your option) any later version.
//**
//**	This program is distributed in the hope that it will be useful,
//**  but WITHOUT ANY WARRANTY; without even the implied warranty of
//**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//**  GNU General Public License for more details.
//**	
//**************************************************************************

// MACROS ------------------------------------------------------------------

#define FLOATSPEED		140.0

// keep track of special lines as they are hit,
// but don't process them until the move is proven valid
#define MAXSPECIALCROSS	32

// CONSTANTS ---------------------------------------------------------------

// TYPES -------------------------------------------------------------------

// FUNCTION PROTOTYPES -----------------------------------------------------

// DATA DECLARATIONS -------------------------------------------------------

float	xspeed[8] = {1.0, 0.7171630859, 0.0, -0.7171630859, -1.0, -0.7171630859, 0.0, 0.7171630859};
float	yspeed[8] = {0.0, 0.7171630859, 1.0, 0.7171630859, 0.0, -0.7171630859, -1.0, -0.7171630859};

boolean			floatok;	// if true, move would be ok if
							// within tmfloorz - tmceilingz
float			saved_tmfloorz;

// keep track of the line that lowers the ceiling,
// so missiles don't explode against sky hack walls
line_t*			ceilingline;

float			tmbbox[4];
mobj_t*			tmthing;
int				tmflags;
float			tmx;
float			tmy;
float			tmz;
TVec			tmpos;
float			tmfloorz;
float			tmceilingz;
float			tmdropoffz;
sec_plane_t*	tmfloor;
sec_plane_t*	tmceiling;

line_t*			spechit[MAXSPECIALCROSS];
int				numspechit;

// CODE --------------------------------------------------------------------

//**************************************************************************
//
//	CHECK RELATIVE POSITION
//
//**************************************************************************

//==========================================================================
//
//	PIT_CheckThing
//
//==========================================================================

boolean PIT_CheckThing(mobj_t* thing)
{
    float		blockdist;
    boolean		solid;
    int			damage;

    if (!(thing->flags & (MF_SOLID | MF_SPECIAL | MF_SHOOTABLE)))
		return true;
    
    blockdist = thing->radius + tmthing->radius;

    if (fabs(thing->origin.x - tmx) >= blockdist || fabs(thing->origin.y - tmy) >= blockdist)
    {
		// didn't hit it
		return true;
    }
    
    // don't clip against self
    if (thing == tmthing)
		return true;

	BlockingMobj = thing;
 	if (tmthing->flags2 & MF2_PASSMOBJ)
	{
		// check if a mobj passed over/under another object
		if (tmthing->type == MT_BISHOP && thing->type == MT_BISHOP)
		{
			// don't let bishops fly over other bishops
			return false;
		}
		if (tmthing->origin.z >= thing->origin.z + thing->height
			&& !(thing->flags & MF_SPECIAL))
		{
			return true;
		}
		else if (tmthing->origin.z + tmthing->height < thing->origin.z
			&& !(thing->flags & MF_SPECIAL))
		{
			// under thing
			return true;
		}
	}
	// Check for skulls slamming into things
	if (tmthing->flags & MF_SKULLFLY)
	{
		if (tmthing->type == MT_MINOTAUR)
		{
			// Slamming minotaurs shouldn't move non-creatures
			if (!(thing->flags & MF_COUNTKILL))
			{
				return false;
			}
		}
		else if (tmthing->type == MT_HOLY_FX)
		{
			if(thing->flags&MF_SHOOTABLE && thing != tmthing->target)
			{
				if (netgame && !deathmatch && thing->player)
				{
					// don't attack other co-op players
					return true;
				}
				if (thing->flags2&MF2_REFLECTIVE
					&& (thing->player || thing->flags2&MF2_BOSS))
				{
					tmthing->specialmobj1 = tmthing->target;
					tmthing->target = thing;
					return true;
				}
				if(thing->flags&MF_COUNTKILL || thing->player)
				{
					tmthing->specialmobj1 = thing;
				}
				if (P_Random() < 96)
				{
					damage = 12;
					if (thing->player || thing->flags2&MF2_BOSS)
					{
						damage = 3;
						// ghost burns out faster when attacking players/bosses
						tmthing->health -= 6;
					}
					DamageMobj(thing, tmthing, tmthing->target, damage);
					if (P_Random() < 128)
					{
						P_SpawnMobj(tmthing->origin.x, tmthing->origin.y, tmthing->origin.z,
							MT_HOLY_PUFF);
						StartSoundName(tmthing, "SpiritAttack");
						if (thing->flags&MF_COUNTKILL && P_Random() < 128
							&& !S_GetSoundPlayingInfo(thing, GetSoundID("PuppyBeat")))
						{
							if ((thing->type == MT_CENTAUR) ||
								(thing->type == MT_CENTAURLEADER) ||
								(thing->type == MT_ETTIN))
							{
								StartSoundName(thing, "PuppyBeat");
							}
						}
					}
				}
				if (thing->health <= 0)
				{
					tmthing->special1 = 0;
					tmthing->specialmobj1 = NULL;
				}
			}
			return true;
		}
		damage = ((P_Random()%8)+1)*tmthing->damage;
		DamageMobj(thing, tmthing, tmthing, damage);
		tmthing->flags &= ~MF_SKULLFLY;
		tmthing->velocity.x = tmthing->velocity.y = tmthing->velocity.z = 0.0;
		SetMobjState(tmthing, tmthing->seestate);
		return false;
	}
	// Check for blasted thing running into another
	if (tmthing->flags2&MF2_BLASTED && thing->flags&MF_SHOOTABLE)
	{
		if (!(thing->flags2&MF2_BOSS) && (thing->flags&MF_COUNTKILL))
		{
			thing->velocity.x += tmthing->velocity.x;
			thing->velocity.y += tmthing->velocity.y;
			if ((thing->velocity.x + thing->velocity.y) > 3.0 * 35.0)
			{
				damage = (ftoi(tmthing->mass)/100)+1;
				DamageMobj(thing, tmthing, tmthing, damage);
				damage = (ftoi(thing->mass)/100)+1;
				DamageMobj(tmthing, thing, thing, damage>>2);
			}
			return false;
		}
	}
	// Check for missile
	if (tmthing->flags&MF_MISSILE)
	{
		// Check for a non-shootable mobj
		if (thing->flags2&MF2_NONSHOOTABLE)
		{
			return true;
		}
		// Check if it went over / under
		if (tmthing->origin.z > thing->origin.z + thing->height)
		{ // Over thing
			return true;
		}
		if (tmthing->origin.z + tmthing->height < thing->origin.z)
		{ // Under thing
			return true;
		}
		if (tmthing->flags2&MF2_FLOORBOUNCE)
		{
			if (tmthing->target == thing || !(thing->flags&MF_SOLID))
			{
				return true;
			}
			else
			{
				return false;
			}
		}
		if (tmthing->type == MT_LIGHTNING_FLOOR
			|| tmthing->type == MT_LIGHTNING_CEILING)
		{
			if (thing->flags&MF_SHOOTABLE && thing != tmthing->target)
			{
				if (thing->mass != 99999.0)
				{
					thing->velocity.x += tmthing->velocity.x / 16.0;
					thing->velocity.y += tmthing->velocity.y / 16.0;
				}
				if((!thing->player && !(thing->flags2&MF2_BOSS))
					|| !(level->tictime&1))
				{
					if(thing->type == MT_CENTAUR
						|| thing->type == MT_CENTAURLEADER)
					{ // Lightning does more damage to centaurs
						DamageMobj(thing, tmthing, tmthing->target, 9);
					}
					else
					{
						DamageMobj(thing, tmthing, tmthing->target, 3);
					}
					if(!(S_GetSoundPlayingInfo(tmthing, 
						GetSoundID("MageLightningZap"))))
					{
						StartSoundName(tmthing, "MageLightningZap");
					}
					if(thing->flags&MF_COUNTKILL && P_Random() < 64
						&& !S_GetSoundPlayingInfo(thing, GetSoundID("PuppyBeat")))
					{
						if ((thing->type == MT_CENTAUR) ||
							(thing->type == MT_CENTAURLEADER) ||
							(thing->type == MT_ETTIN))
						{
							StartSoundName(thing, "PuppyBeat");
						}
					}
				}
				tmthing->health--;
				if(tmthing->health <= 0 || thing->health <= 0)
				{
					return false;
				}
				if (tmthing->type == MT_LIGHTNING_FLOOR)
				{
					if (tmthing->specialmobj2 &&
						!tmthing->specialmobj2->specialmobj1)
					{
						tmthing->specialmobj2->specialmobj1 = thing;
					}
				}
				else if (!tmthing->specialmobj1)
				{
					tmthing->specialmobj1 = thing;
				}
			}
			return true; // lightning zaps through all sprites
		}
		else if(tmthing->type == MT_LIGHTNING_ZAP)
		{
			mobj_t *lmo;

			if (thing->flags & MF_SHOOTABLE && thing != tmthing->target)
			{			
				lmo = tmthing->specialmobj2;
				if (lmo)
				{
					if (lmo->type == MT_LIGHTNING_FLOOR)
					{
						if (lmo->specialmobj2 &&
							!lmo->specialmobj2->specialmobj1)
						{
							lmo->specialmobj2->specialmobj1 = thing;
						}
					}
					else if (!lmo->specialmobj1)
					{
						lmo->specialmobj1 = thing;
					}
					if (!(level->tictime & 3))
					{
						lmo->health--;
					}
				}
			}
		}
		else if (tmthing->type == MT_MSTAFF_FX2 && thing != tmthing->target)
		{
			if (!thing->player && !(thing->flags2&MF2_BOSS))
			{
				switch (thing->type)
				{
					case MT_FIGHTER_BOSS:	// these not flagged boss
					case MT_CLERIC_BOSS:	// so they can be blasted
					case MT_MAGE_BOSS:
						break;
					default:
						DamageMobj(thing, tmthing, tmthing->target, 10);
						return true;
						break;
				}
			}
		}
		if (tmthing->target && tmthing->target->type == thing->type)
		{
			// Don't hit same species as originator
			if (thing == tmthing->target)
			{
				// Don't missile self
				return true;
			}
			if (!thing->player)
			{
				// Hit same species as originator, explode, no damage
				return false;
			}
		}
		if (!(thing->flags&MF_SHOOTABLE))
		{
			// Didn't do any damage
			return !(thing->flags & MF_SOLID);
		}
		if (tmthing->flags2&MF2_RIP)
		{
			if (!(thing->flags&MF_NOBLOOD) &&
				!(thing->flags2&MF2_REFLECTIVE) &&
				!(thing->flags2&MF2_INVULNERABLE))
			{ // Ok to spawn some blood
				P_RipperBlood(tmthing);
			}
			//StartSoundName(tmthing, sfx_ripslop);
			damage = ((P_Random()&3)+2)*tmthing->damage;
			DamageMobj(thing, tmthing, tmthing->target, damage);
			if(thing->flags2&MF2_PUSHABLE
				&& !(tmthing->flags2&MF2_CANNOTPUSH))
			{ // Push thing
				thing->velocity.x += tmthing->velocity.x / 4.0;
				thing->velocity.y += tmthing->velocity.y / 4.0;
			}
			numspechit = 0;
			return true;
		}
		// Do damage
		damage = ((P_Random()%8)+1)*tmthing->damage;
		if (damage)
		{
			if (!(thing->flags&MF_NOBLOOD) && 
				!(thing->flags2&MF2_REFLECTIVE) &&
				!(thing->flags2&MF2_INVULNERABLE) &&
				!(tmthing->type == MT_TELOTHER_FX1) &&
				!(tmthing->type == MT_TELOTHER_FX2) &&
				!(tmthing->type == MT_TELOTHER_FX3) &&
				!(tmthing->type == MT_TELOTHER_FX4) &&
				!(tmthing->type == MT_TELOTHER_FX5) &&
				(P_Random() < 192))
			{
				P_BloodSplatter(tmthing->origin.x, tmthing->origin.y, tmthing->origin.z, thing);
			}
			DamageMobj(thing, tmthing, tmthing->target, damage);
		}
		return false;
	}
	if (thing->flags2&MF2_PUSHABLE && !(tmthing->flags2&MF2_CANNOTPUSH))
	{
		// Push thing
		thing->velocity.x += tmthing->velocity.x / 4.0;
		thing->velocity.y += tmthing->velocity.y / 4.0;
	}
	// Check for special thing
	if (thing->flags&MF_SPECIAL)
	{
		solid = thing->flags&MF_SOLID;
		if (tmthing->flags&MF_PICKUP)
		{
			// Can be picked up by tmthing
			P_TouchSpecialThing(thing, tmthing); // Can remove thing
		}
		return !solid;
	}
	return !(thing->flags & MF_SOLID);
}

//===========================================================================
//
// CheckForPushSpecial
//
//===========================================================================

void CheckForPushSpecial(line_t *line, int side, mobj_t *mobj)
{
	if (line->special)
	{
		if (mobj->flags2 & MF2_PUSHWALL)
		{
			P_ActivateLine(line, mobj, side, SPAC_PUSH);
		}
		else if (mobj->flags2 & MF2_IMPACT)
		{
			P_ActivateLine(line, mobj, side, SPAC_IMPACT);
		}	
	}
}

//==========================================================================
//
//	PIT_CheckLine
//
//	Adjusts tmfloorz and tmceilingz as lines are contacted
//
//==========================================================================

boolean PIT_CheckLine(line_t* ld)
{
	TVec		hit_point;
	opening_t	*open;

    if (tmbbox[BOXRIGHT] <= ld->bbox[BOXLEFT] ||
    	tmbbox[BOXLEFT] >= ld->bbox[BOXRIGHT] ||
    	tmbbox[BOXTOP] <= ld->bbox[BOXBOTTOM] ||
    	tmbbox[BOXBOTTOM] >= ld->bbox[BOXTOP])
		return true;

	if (P_BoxOnLineSide(&tmbbox[0], ld) != -1)
		return true;
		
    // A line has been hit
    
    // The moving thing's destination position will cross
    // the given line.
    // If this should not be allowed, return false.
    // If the line is special, keep track of it
    // to process later if the move is proven ok.
    // NOTE: specials are NOT sorted by order,
    // so two special lines that are only 8 pixels apart
    // could be crossed in either order.
    
    if (!ld->backsector)
	{
		// One sided line
		if (tmthing->flags2 & MF2_BLASTED)
		{
			DamageMobj(tmthing, NULL, NULL, ftoi(tmthing->mass) >> 5);
		}
		CheckForPushSpecial(ld, 0, tmthing);
		return false;
	}

    if (!(tmthing->flags & MF_MISSILE))
    {
		if ( ld->flags & ML_BLOCKING )
		{
			// Explicitly blocking everything
			if (tmthing->flags2 & MF2_BLASTED)
			{
				DamageMobj(tmthing, NULL, NULL, ftoi(tmthing->mass) >> 5);
			}
			CheckForPushSpecial(ld, 0, tmthing);
			return false;
		}

		if (!tmthing->player && ld->flags & ML_BLOCKMONSTERS)
		{
			// Block monsters only
			if (tmthing->flags2 & MF2_BLASTED)
			{
				DamageMobj(tmthing, NULL, NULL, ftoi(tmthing->mass) >> 5);
			}
			return false;
		}
    }

    // set openrange, opentop, openbottom
	hit_point = tmpos - (DotProduct(tmpos, ld->normal) - ld->dist) * ld->normal;
	open = LineOpenings(ld, hit_point);
	open = FindOpening(open, tmz, tmz + tmthing->height);

	if (open)
	{
	    // adjust floor / ceiling heights
    	if (!(open->ceiling->flags & SPF_NOBLOCKING) && open->top < tmceilingz)
	    {
			tmceiling = open->ceiling;
			tmceilingz = open->top;
			ceilingline = ld;
	    }

	    if (!(open->floor->flags & SPF_NOBLOCKING) && open->bottom > tmfloorz)
		{
			tmfloor = open->floor;
			tmfloorz = open->bottom;
		}

	    if (open->lowfloor < tmdropoffz)
			tmdropoffz = open->lowfloor;
	}
	else
	{
		tmceilingz = tmfloorz;
	}
		
    // if contacted a special line, add it to the list
    if (ld->special)
    {
    	if (numspechit >= MAXSPECIALCROSS)
        {
//        	cond << "Spechit overflow\n";
		}
        else
        {
			spechit[numspechit] = ld;
			numspechit++;
		}
    }

    return true;
}

//==========================================================================
//
//	P_CheckPosition
//
//	This is purely informative, nothing is modified
// (except things picked up).
// 
// in:
//  a mobj_t (can be valid or invalid)
//  a position to be checked
//   (doesn't need to be related to the mobj_t->x,y)
//
// during:
//  special things are touched if MF_PICKUP
//  early out on solid lines?
//
// out:
//  newsubsec
//  floorz
//  ceilingz
//  tmdropoffz
//   the lowest point contacted
//   (monsters won't move to a dropoff)
//  speciallines[]
//  numspeciallines
//
//==========================================================================

boolean P_CheckPosition(mobj_t* thing, float x, float y)
{
    int				xl;
    int				xh;
    int				yl;
    int				yh;
    int				bx;
    int				by;
    subsector_t*	newsubsec;
	sec_region_t	*gap;
	sec_region_t	*reg;

    tmthing = thing;
    tmflags = thing->flags;
	
    tmx = x;
    tmy = y;
	tmz = thing->origin.z;
	tmpos = vector(x, y, thing->origin.z);
	
    tmbbox[BOXTOP] = y + tmthing->radius;
    tmbbox[BOXBOTTOM] = y - tmthing->radius;
    tmbbox[BOXRIGHT] = x + tmthing->radius;
    tmbbox[BOXLEFT] = x - tmthing->radius;

    newsubsec = PointInSubsector(x, y);
    ceilingline = NULL;
    
    // The base floor / ceiling is from the subsector
    // that contains the point.
    // Any contacted lines the step closer together
    // will adjust them.
	gap = FindThingGap(newsubsec->sector->botregion, tmpos, tmz, tmz + thing->height);
	reg = gap;
	while (reg->prev && reg->floor->flags & SPF_NOBLOCKING)
		reg = reg->prev;
    tmfloor = reg->floor;
    tmfloorz = tmdropoffz = GetPlanePointZ(reg->floor, tmpos);
	reg = gap;
	while (reg->next && reg->ceiling->flags & SPF_NOBLOCKING)
		reg = reg->next;
    tmceiling = reg->ceiling;
    tmceilingz = GetPlanePointZ(reg->ceiling, tmpos);

    (*validcount)++;
    numspechit = 0;

	if (tmflags & MF_NOCLIP && !(tmflags & MF_SKULLFLY))
		return true;
	BlockingMobj = NULL;
    
    // Check things first, possibly picking things up.
    // The bounding box is extended by MAXRADIUS
    // because mobj_ts are grouped into mapblocks
    // based on their origin point, and can overlap
    // into adjacent blocks by up to MAXRADIUS units.
    xl = MapBlock(tmbbox[BOXLEFT] - level->bmaporgx - MAXRADIUS);
    xh = MapBlock(tmbbox[BOXRIGHT] - level->bmaporgx + MAXRADIUS);
    yl = MapBlock(tmbbox[BOXBOTTOM] - level->bmaporgy - MAXRADIUS);
    yh = MapBlock(tmbbox[BOXTOP] - level->bmaporgy + MAXRADIUS);

    for (bx = xl; bx <= xh; bx++)
		for (by = yl; by <= yh; by++)
	    	if (!P_BlockThingsIterator(bx, by, PIT_CheckThing))
				return false;
    
    // check lines
	if (tmflags & MF_NOCLIP)
	{
		return true;
	}
	BlockingMobj = NULL;

    xl = MapBlock(tmbbox[BOXLEFT] - level->bmaporgx);
    xh = MapBlock(tmbbox[BOXRIGHT] - level->bmaporgx);
    yl = MapBlock(tmbbox[BOXBOTTOM] - level->bmaporgy);
    yh = MapBlock(tmbbox[BOXTOP] - level->bmaporgy);

    for (bx = xl; bx <= xh; bx++)
		for (by = yl; by <= yh; by++)
	    	if (!P_BlockLinesIterator(bx, by, PIT_CheckLine))
				return false;

    return true;
}

//==========================================================================
//
//  PushLine
//
//==========================================================================

void PushLine(mobj_t *thing)
{
	if (!(thing->flags & (MF_TELEPORT | MF_NOCLIP)))
	{
		int			numSpecHitTemp;
		line_t		*ld;
		int			side;

		if (thing->flags2 & MF2_BLASTED)
		{
			DamageMobj(thing, NULL, NULL, ftoi(thing->mass) >> 5);
		}
		numSpecHitTemp = numspechit;
		while (numSpecHitTemp > 0)
		{
			numSpecHitTemp--;
			// see if the line was crossed
			ld = spechit[numSpecHitTemp];
			side = PointOnPlaneSide(thing->origin, ld);
			CheckForPushSpecial(ld, side, thing);
		}
	}
}

//==========================================================================
//
//	P_TryMove
//
//	Attempt to move to a new position, crossing special lines unless
// MF_TELEPORT is set.
//
//==========================================================================

boolean P_TryMove(mobj_t* thing, float x, float y)
{
	boolean		check;
	TVec		oldorg;
	int			side;
	int			oldside;
	line_t		*ld;

	check = P_CheckPosition(thing, x, y);
    floatok = false;
	saved_tmfloorz = tmfloorz;
	if (!check)
	{
		if (!BlockingMobj || BlockingMobj->player || !thing->player ||
			BlockingMobj->origin.z + BlockingMobj->height - thing->origin.z > 24.0 ||
			(BlockingMobj->subsector->sector->ceilingheight - (BlockingMobj->origin.z + BlockingMobj->height) < thing->height) ||
			(tmceilingz - (BlockingMobj->origin.z + BlockingMobj->height) < thing->height)
			)
		{
			PushLine(thing);
			return false;
		}
	}
    
    if (!(thing->flags & MF_NOCLIP))
    {
		if (tmceilingz - tmfloorz < thing->height)
		{
			// Doesn't fit
			PushLine(thing);
			return false;
		}

		floatok = true;
	
		if (!(thing->flags & MF_TELEPORT) &&
			tmceilingz - thing->origin.z < thing->height &&
			thing->type != MT_LIGHTNING_CEILING &&
			!(thing->flags2 & MF2_FLY))
		{
			// mobj must lower itself to fit
			PushLine(thing);
			return false;
		}
		if (thing->flags2 & MF2_FLY)
		{
			if (thing->origin.z + thing->height > tmceilingz)
			{
				thing->velocity.z = -8.0 * 35.0;
				PushLine(thing);
				return false;
			}
			else if (thing->origin.z < tmfloorz && tmfloorz - tmdropoffz > 24.0)
			{
				thing->velocity.z = 8.0 * 35.0;
				PushLine(thing);
				return false;
			}
		}
		if (!(thing->flags & MF_TELEPORT)
			// The Minotaur floor fire (MT_MNTRFX2) can step up any amount
			&& thing->type != MT_MNTRFX2 && thing->type != MT_LIGHTNING_FLOOR
			&& tmfloorz - thing->origin.z > 24.0)
		{
			PushLine(thing);
			return false;
		}
		if (!(thing->flags & (MF_DROPOFF | MF_FLOAT)) &&
			(tmfloorz - tmdropoffz > 24.0) &&
			!(thing->flags2 & MF2_BLASTED))
		{
			// Can't move over a dropoff unless it's been blasted
				return false;
		}
		if (thing->flags2 & MF2_CANTLEAVEFLOORPIC &&
			(tmfloor->pic != thing->floor->pic
				|| tmfloorz != thing->origin.z))
		{
			// must stay within a sector of a certain floor type
			return false;
		}
    }
    
    // the move is ok,
    // so link the thing into its new position
    P_UnsetThingPosition(thing);

    oldorg = thing->origin;
    thing->origin.x = x;
    thing->origin.y = y;

	P_SetThingPosition(thing);
	thing->floor = tmfloor;
	thing->ceiling = tmceiling;
	thing->floorz = tmfloorz;
	thing->ceilingz = tmceilingz;

	if (thing->flags2 & MF2_FLOORCLIP)
	{
		if (thing->origin.z == thing->subsector->sector->floorheight &&
			P_GetThingFloorType(thing) >= FLOOR_LIQUID)
		{
			thing->floorclip = 10.0;
		}
		else 
		{
			thing->floorclip = 0.0;
		}
	}

	//
	// if any special lines were hit, do the effect
	//
	if (!(thing->flags & (MF_TELEPORT | MF_NOCLIP)))
	{
		while (numspechit > 0)
		{
			numspechit--;
		    // see if the line was crossed
		    ld = spechit[numspechit];
		    side = PointOnPlaneSide(thing->origin, ld);
	    	oldside = PointOnPlaneSide(oldorg, ld);
		    if (side != oldside)
		    {
				if (ld->special)
				{
					if (thing->player)
					{
						P_ActivateLine(ld, thing, oldside, SPAC_CROSS);
					}
					else if (thing->flags2 & MF2_MCROSS)
					{
						P_ActivateLine(ld, thing, oldside, SPAC_MCROSS);
					}
					else if (thing->flags2 & MF2_PCROSS)
					{
						P_ActivateLine(ld, thing, oldside, SPAC_PCROSS);
					}
				}
			}
		}
	}

	return true;
}

//==========================================================================
//
//	P_Move
//
//	Move in the current direction, returns false if the move is blocked.
//
//==========================================================================

boolean P_Move(mobj_t*	actor)
{
    float		tryx;
    float		tryy;
    
    line_t*		ld;
    boolean		good;
    
	if (actor->flags2 & MF2_BLASTED) return true;
    if (actor->movedir == DI_NODIR)
		return false;

#ifdef RANGECHECK
    if (actor->movedir >= 8 || actor->movedir < 0)
		I_Error("Weird actor->movedir!");
#endif
		
    tryx = actor->origin.x + actor->speed * frametime * xspeed[actor->movedir];
    tryy = actor->origin.y + actor->speed * frametime * yspeed[actor->movedir];

    if (!P_TryMove(actor, tryx, tryy))
    {
		// open any specials
		if (actor->flags & MF_FLOAT && floatok)
		{
	    	// must adjust height
	    	if (actor->origin.z < saved_tmfloorz)
				actor->origin.z += FLOATSPEED * frametime;
	    	else
				actor->origin.z -= FLOATSPEED * frametime;

	    	actor->flags |= MF_INFLOAT;
	    	return true;
		}
		
		if (!numspechit)
	    	return false;
			
		actor->movedir = DI_NODIR;
		good = false;
		while (numspechit--)
		{
	    	ld = spechit[numspechit];
	    	// if the special is not a door
	    	// that can be opened,
	    	// return false
	    	if (P_ActivateLine(ld, actor, 0, SPAC_USE))
			{
				good = true;
			}
		}
		return good;
    }
    else
    {
		actor->flags &= ~MF_INFLOAT;
    }
	
	if (!(actor->flags&MF_FLOAT))
	{
		if (actor->origin.z > actor->floorz)
		{
			P_HitFloor(actor);
		}
		actor->origin.z = actor->floorz;
	}
    return true; 
}

